-- details:
-- DONE: Amy jump height (jumpfactor in S_SKIN lump)
local skin = "honey"

-- Resets all of the player's fields..
-- Also used to instantiate everything within PlayerSpawn
-- (some exceptions)
local resetFields = function(player)
	player.mo.deathtype = 0 			-- Used for deaths
	
	-- animation variables
	player.mo.aekickstate = 0			-- indicates whether to do the aesthetic kick or not
	player.mo.aeghost = 0				-- indicates how long afterimage is active
	
	player.mo.collideanimtime = 0
	-- TODO: create variable which indicates the delay between frames for animation?
	player.mo.animreset = true 		-- Indicates when player state animation should be reset (typically after doing things like crouching, etc.)
	
	-- idle animation
	player.mo.idletimer = 0 				-- Timer that increments; it is used to change between player's idle state frames.

	-- crouching variables
	player.mo.crouchtrue = true			-- indicates if player is crouching or not
	player.mo.uncrouch = false			-- indicates that the default state should be restored once
	
--------------Attacks
	-- ground quake variables
	player.mo.fastfalltrue = true
	player.mo.fastfallsfx = true
	player.mo.fastfalltime = 0
	player.mo.fastfallquake = false 	-- indicates if player can perform AOE quake
	
	-- Dive variables
	player.mo.divetrue = true 			-- indicates if player is currently diving
	player.mo.divetimer = 0				-- indicates for how long the dive is active
	player.mo.divestate = true			-- Used to enter the dive state (only ONCE)
	player.mo.divesfx = false			-- to play sfx (only ONCE)
	-- Roll variables
	player.mo.rolltrue = true
	player.mo.rolltimer = 0
	player.mo.rollstate = true
	player.mo.rollsfx = false
	--
	player.mo.c1smomstop = true 			-- used to kill momentum upon landing
			
	-- backflip variables
	player.mo.backfliptrue = true		-- indicates if player is currently backflipping
	player.mo.backfliptimer = 0 		-- Indicates for how long the backflip is active
	player.mo.unbackflip = true			-- Used to enter the backflip state (only ONCE!!)
	player.mo.backflipsfx = true 		-- to play sfx only once
	
	-- airgrab variables
	player.mo.airgrabattempt = true		-- indicates if the player is currently attempting to air grab
	player.mo.airgrabtrue = true 		-- indicates if player is currently airgrabbing
--End of attacks-----------------------
	
	-- nocontrol variables
	player.mo.nocontrol = false 		-- might be unused, indicates if player has no control at the moment
	
	-- player's rollangle
	player.mo.rollangle = 0
end

-- Initialize player's variables
addHook("PlayerSpawn", function(player)
	resetFields(player)
end)

-- Sanity check: Make sure the player IS a player
local sanityCheck = function(player)
	if (not player.mo or not player.mo.valid) then
		return false
	else
		return true
	end
end

-- TODO: Honey's ground quake. Prerequisites: falling from sufficient height turns a variable on, which then allows you to do the quake
-- otherwise, you don't do a quake. Though do checks to make sure that on the way down the momentum doesn't reset.. 
-- i.e. if momz >= limit, then quake = true elseif momz < limit then quake = false
--sha5


-- TODO: currentlyAttacking conditions for all moves: The reasoning behind this is to allow player to perform only specific key combinations
-- Take away control from player during one of these states
-- (This is used only in the AbilitySpecial attack so far)
local currentlyActing = function(player)
	if	not player.mo.crouchtrue   or
		not player.mo.fastfalltrue or
		not player.mo.divetrue 	   or
		not player.mo.backfliptrue or
		not player.mo.airgrabtrue
	then
		return true
	else
		return false
	end
end
-- PreThinkFrame is best used for handling inputs
addHook("PreThinkFrame", function()
	for player in players.iterate do
		if sanityCheck(player) and player.mo.skin ~= skin then
			return
		end
	
		-- (Reset AbilitySpecial variables)
		if not((player.cmd.buttons & BT_JUMP) and (player.mo.fastfalltrue)) and P_IsObjectOnGround(player.mo) then
			player.mo.fastfalltrue = true
			player.mo.fastfalltime = 0
			player.mo.fastfallsfx = true
		end
	
		-- Jump: Normally attacks like sonic/tails/etc. Has aesthetic kick like bandages

---------- Crouching: C1
		if (player.cmd.buttons & BT_CUSTOM1) and P_IsObjectOnGround(player.mo) --and not currentlyAttacking(player) then
			player.mo.crouchtrue = false
		end
	
		if not(player.cmd.buttons & BT_CUSTOM1) and not player.mo.crouchtrue then
			player.mo.crouchtrue = true
		end
-- Crouching end

---------- Dive: C1+Spin
		if (player.cmd.buttons & BT_CUSTOM1) and (player.cmd.buttons & BT_USE) and player.mo.divetrue and P_IsObjectOnGround(player.mo)-- and not currentlyAttacking(player) then
			player.mo.crouchtrue = true 	-- disable crouching
			player.mo.divetrue = false  	-- enable dive
			--P_SetObjectMomZ(player.mo, 2, false)
			P_InstaThrust(player.mo, player.drawangle, 17*FRACUNIT)
			P_SetObjectMomZ(player.mo, 3*FRACUNIT, false)
			print("Dive")
		end
		if not player.mo.divetrue then
			player.mo.divetimer = $ + 1
			if not P_IsObjectOnGround(player.mo) then
				player.mo.c1smomstop = false
			end
		end
		if player.mo.divetimer >= 20 then
			-- Reset everything
			player.mo.divetrue = true
			player.mo.divesfx = true
			player.mo.divetimer = 0
			player.mo.divestate = true
			-- If crouch is still held then we enter crouch state
			if player.cmd.buttons & BT_CUSTOM1 and P_IsObjectOnGround(player.mo) then
				player.mo.state = S_PLAY_HCROUCH
			elseif P_IsObjectOnGround(player.mo) then
				player.mo.state = S_PLAY_STND
			end
		end
		if P_IsObjectOnGround(player.mo) and not player.mo.c1smomstop then
			-- Turn off the switch
			player.mo.c1smomstop = true 
			-- Kill all momentum
			player.mo.momx = 0
			player.mo.momy = 0
			player.mo.momz = 0
			-- Kill timer
			player.mo.divetimer = 20
		end
		-- Things to do while dive is active
		if not player.mo.divetrue then
			-- Change to dive state
			if player.mo.divestate and not P_IsObjectOnGround(player.mo) then
				player.mo.state = S_PLAY_HDIVE1
				player.mo.divestate = false
			end
		end
--Dive end

---------- Roll: C1+spin + [during opportunity window] spin
		if (player.cmd.buttons & BT_CUSTOM1) and (player.cmd.buttons & BT_USE) and player.mo.divetrue and not P_IsObjectOnGround(player.mo)
			player.mo.crouchtrue = true 	-- disable crouching
			player.mo.divetrue = false  	-- enable dive
			--P_SetObjectMomZ(player.mo, 2, false)
			P_InstaThrust(player.mo, player.drawangle, 17*FRACUNIT)
			P_SetObjectMomZ(player.mo, 3*FRACUNIT, false)
			print("Dive")
		end
-- Roll end

-- TODO: slightly influence direction while rising (how can i do this?)
---------- Backflip: C1+Jump
		if (player.cmd.buttons & BT_CUSTOM1) and (player.cmd.buttons & BT_JUMP) and player.mo.backfliptrue and P_IsObjectOnGround(player.mo) then
			player.mo.crouchtrue = true 	-- disable crouching
			player.mo.backfliptrue = false 	-- enable backflip
			--player.pflags = $ & ~PF_SPINNING -- disable spinning
			P_SetObjectMomZ(player.mo, 28*FRACUNIT, true) --false?
		end
		if not player.mo.backfliptrue then
			player.mo.backfliptimer = $ + 1 --increment backflip timer: indicates how long we are in an attack state
		end
		if player.mo.backfliptimer >= 20 then
			player.mo.backfliptrue = true	-- Reset backflip variable
			player.mo.backflipsfx = true	-- Reset backflip sfx
			player.mo.backfliptimer = 0 	-- Reset backflip timer
			player.mo.unbackflip = true 	-- Reset unbackflip
			player.mo.state = S_PLAY_FALL	-- player is now falling
		end
		-- Things to do while backflipping is active
		if not player.mo.backfliptrue then
			-- Play backflip sfx
			if player.mo.backflipsfx then
				S_StartSound(player.mo, sfx_s3k6b)
				player.mo.backflipsfx = false
			end
			-- Change to the backflip state
			if player.mo.unbackflip and not P_IsObjectOnGround(player.mo) then
				player.mo.state = S_PLAY_HBKFLP1
				player.mo.unbackflip = false
			end
			-- Slow down on the peak
			-- TODO: clean this up
			local mo = player.mo
			local friction = FRACUNIT
			local zfriction = FRACUNIT*9/10
			local zlimit = nil
			if not P_IsObjectOnGround(player.mo) then
				local limit = FRACUNIT*80--FixedMul(mo.scale,$)
				local spd = min(limit,max(-limit,
					FixedHypot(mo.momx,mo.momy)
				))
		-- 		friction = FixedDiv($,mo.scale)
				spd = FixedMul(spd,friction)
				local ang = R_PointToAngle2(0,0,mo.momx,mo.momy)
				mo.momx = P_ReturnThrustX(nil,ang,spd)
				mo.momy = P_ReturnThrustY(nil,ang,spd)
				//Z movement
				if zfriction == nil then zfriction = friction end
				if zlimit == nil then zlimit = limit end
				zlimit = FixedDiv($,mo.scale)
				mo.momz = min(zlimit,max(-zlimit,
					FixedMul($,zfriction)
				))
			end
		end
		if player.mo.backfliptimer == 8 then
			P_InstaThrust(player.mo, player.drawangle, -5*FRACUNIT)
		end
		if player.mo.backfliptimer >= 5 and player.mo.backfliptimer <= 7 then
			P_SetObjectMomZ(player.mo, player.mo.momz, false)
		end
--Backflip End

--print(R_PointToAngle2(0,0,player.cmd.forwardmove,player.cmd.sidemove))
	
	
---------- Tech: C1 + Direction (while in Pain state)
--Tech End
	
		-- Air grab: Jump then Spin
		-- Small air stall (moves slowly in the air), if comes in contact with someone then throws
		if (player.cmd.buttons & BT_USE) and (player.cmd.buttons & BT_JUMP) and not P_IsObjectOnGround(player.mo) and player.mo.airgrabtrue-- and not currentlyAttacking(player) then
			print("Air grab")
			player.mo.airgrabtrue = false
		end
	
		if not((player.cmd.buttons & BT_USE) and (player.cmd.buttons & BT_JUMP)) and not player.mo.airgrabtrue then
			player.mo.airgrabtrue = true
		end
	
		-- Halt player's movement if they're in a specific state/action
		if currentlyActing(player) then
			player.cmd.sidemove = 0
			player.cmd.forwardmove = 0
		end
	end
end)

--======Credit: Bloopsdoob, Lach=======
addHook("ThinkFrame", function()
	for p in players.iterate do
		if sanityCheck(p) and p.mo.skin == skin then 	-- I only edited this
            if p.playerstate != PST_LIVE
                if p.mo.deathtype == 1
                    p.mo.state = S_PLAY_HONEY_DIE_FIRE
                    p.mo.frame = C
                    p.mo.deathtype = 0
                elseif p.mo.deathtype == 2
                    p.mo.state = S_PLAY_HONEY_DIE_ELEC
                    p.mo.deathtype = 0
                elseif p.mo.deathtype == 3
                    p.mo.state = S_PLAY_HONEY_DIE_NUKE
                    p.mo.frame = D
                    --dont reset it immediately
                end
                local r = p.mo.radius/FRACUNIT
                local xy = (P_RandomRange(r, -r)*FRACUNIT)
                local z = (P_RandomKey(p.mo.height/FRACUNIT)*FRACUNIT)
                if p.mo.state == S_PLAY_HONEY_DIE_FIRE --burned cat, smokey
                    P_SpawnMobjFromMobj(p.mo, xy, xy, z, MT_SMOKE)
                elseif p.mo.state == S_PLAY_HONEY_DIE_ELEC --zappy cat, still static
                    if p.mo.frame == C --to explain, they both use the same spr2 set, this *would* advance into the fire frame, this prevents so.
                        p.mo.state = S_PLAY_HONEY_DIE_ELEC
                    end

                    if leveltime % 2 == 0
                        local zap = P_SpawnMobjFromMobj(p.mo, xy, xy, z, MT_WATERZAP)
                        zap.colorized = true
                        zap.color = SKINCOLOR_GOLD
                    end
                elseif p.mo.state == S_PLAY_HONEY_DIE_NUKE
                    if p.mo.deathtype == 3 --only runs once before it gets reset.
                        local clothframes = {A, B, B, C, D, E, E, F, F, G, G, H, H, I, I, I, I, I, I, J, J, K, K}
                        for i = 1, 23
                            local cloth = P_SpawnMobjFromMobj(p.mo, 0, 0, p.mo.height/2, MT_HONEY_NUKE_GIB)
                            cloth.fuse = 10*TICRATE
                            cloth.color = p.mo.color
                            cloth.skin = "honey"
                            cloth.state = S_PLAY_HONEY_DIE_NUKE_GIB
                            cloth.frame = clothframes[i]
                            P_InstaThrust(cloth, FixedAngle(P_RandomRange(1,359)*FRACUNIT), P_RandomRange(1,6)*FRACUNIT)
                            P_SetObjectMomZ(cloth, P_RandomRange(3,12)*FRACUNIT, false)
                        end
                        p.mo.deathtype = 0
                    end
                end
            end
        end
--===========================
		local player = p
		-- If you fastfall into a spring, you perform a powerspring
		if player.mo.eflags & MFE_SPRUNG and not player.mo.fastfalltrue then
			--TODO: are these apropriate heights?
			player.mo.momz = $*5/4
			player.mo.momx = $*5/4
			player.mo.momy = $*5/4
			-- We hit a spring during fastfall, so reset this flag and allow player to move freely
			player.mo.fastfalltrue = true
			S_StartSound(player.mo, sfx_sprong)
		end
		
		local sectorthing = P_PlayerTouchingSectorSpecial(player, 0, 0)
		if sectorthing.valid then
			for fof in sectorthing.ffloors()
				if  fof.valid and (fof.flags & FF_BUSTUP) and (fof.flags & FF_EXISTS) and
				( -- Wall break on the floor
					(abs(fof.topheight - player.mo.z) <= 30*player.mo.scale)
					-- If player is currently fastfalling
					and not player.mo.fastfalltrue
				)
					or
				( -- Wall break on the ceiling
					(abs(fof.bottomheight - (player.mo.z + player.mo.height)) <= 30*player.mo.scale)
					-- If player is currently backflipping
					and not player.mo.backfliptrue
				)
				then
					EV_CrumbleChain(player.sectorthing, fof)
				end
			end
		end
	end
end, MT_PLAYER)

-- Fires every time jump is pressed again in the air
-- Hook used to perform special air ability and determine whether doing special air ability in the air should be disabled or enabled
addHook("AbilitySpecial", function(player)
	if sanityCheck(player) and player.mo.skin ~= skin then
		return
	end
	
	-- Take away controls if currently attacking
	--if currentlyAttacking(player) then
	--	player.mo.nocontrol = true
	--	return false
	--end
	
	-- Jump + Jump: fastfall to ground, small AOE quake
	-- TODO: perform `delay` upon landing, probably play a small `delay` animation as well?
	-- TODO: Will not bounce off enemies
	if (player.cmd.buttons & BT_JUMP) and player.mo.fastfalltrue and not currentlyActing(player) then
		-- activate fastfall
		player.mo.fastfalltrue = false
		-- activate quake fastfall (only if momZ is sufficient, look at TODO)
		-- Reset collision animation completely
		-- TODO: should we check for it, or just do it this way
		player.mo.aekickstate = 0
		--player.mo.collideanimtime = 0
		-- TOOD: print("Shoot down, AOE quake") when from sufficient height (sufficient momz)
	end
	return true
end)

-- Used to determine if player can spin/jump
local function determineNoControl(player)
	if sanityCheck(player) and player.mo.skin ~= skin then
		return false
	end
	
	-- TODO: make backflip and roll attacks
	-- Overriding the default behaviour: unable to do anything
	if currentlyActing(player) then
		return true
	end
end
-- Hook used to determine whether spin should be disabled or enabled
-- Also used for spin attack
addHook("SpinSpecial", function(player)
	if determineNoControl(player) then 
		return true 
	end
end)
-- Hook used to determine whether jump should be disabled or enabled
addHook("JumpSpecial", function(player) if determineNoControl(player) then return true end end)

-- Using this to reset variables when:
-- 1. In pain state
-- TODO: should we use this hook to perform various things like thrusts(modify momentums), etc.?
-- NOTE: PostThinkFrame is already used to perform animations/SFX!!
addHook("PlayerThink", function(player)
	if sanityCheck(player) and player.mo.skin ~= skin then
		return
	end
	
	if P_PlayerInPain(player) then
		resetFields(player)
	end
end)

-- Horizontal Wall Breaker Function: Somersault destroys walls
local breakWall = function(mobj, sectorthing)
	local nerds = false
	for fof in sectorthing.ffloors()
		if fof.valid and (fof.flags & FF_BUSTUP) and (fof.flags & FF_EXISTS)
		and fof.topheight >= mobj.z+mobj.height and fof.bottomheight <= mobj.z
		and not mobj.backfliptrue -- player must be doing the backflip kick
		and mobj.skin == skin
		then
			EV_CrumbleChain(sectorthing, fof)
			nerds = true
		end
	end
	return nerds
end

-- Horizontal Wall Breaker (uses hook)
addHook("MobjLineCollide", function(mobj, line)
	if not (line and line.valid and mobj and mobj.valid) and mobj.skin ~= skin then
		return 
	end
	if line.frontsector and line.frontsector.valid then
		breakWall(mobj, line.frontsector) 
	end
	if line.backsector and line.backsector.valid then 
		breakWall(mobj, line.backsector) 
	end
end)

addHook("PlayerCanDamage", function(player, mobj)
	--TODO: Was suppose to use sanityCheck but using manual sanity check instead
	if player.mo.skin ~= skin then
		return
	end
	
	if 	player.pflags & PF_JUMPED 
		and (abs(P_AproxDistance(mobj.x - player.mo.x, mobj.y - player.mo.y)) <= (player.mo.radius+mobj.radius)*2)
		and (mobj.z - mobj.scale <= player.mo.z + player.mo.height + player.mo.momz
		and mobj.z + mobj.height + mobj.scale >= player.mo.z + player.mo.momz)
	then
		-- Since collision with something occurred, set animation variable to be true
		player.mo.aekickstate = 1
		-- Face the direction the target was facing when you hit it
		player.drawangle = R_PointToAngle2(player.mo.x, player.mo.y, mobj.x, mobj.y)
		return true
	end
end)--[int objecttype])

-- Are we dealing with real objects here?
local sanityThingcheck = function(thing, tmthing)
	if not (thing and thing.valid and tmthing and tmthing.valid) 
		return true
	else
		return false
	end
end

-- TODO: Match player's z /height with the spikes, because if you're passing above them while fastfall is true then you'll break them despite not touching them!
local spikeDestroyer = function(thing, tmthing)
	if sanityThingcheck(thing, tmthing) and tmthing.player.skin ~= skin then
		return
	end
	
	-- If fastfall is active then destroy spike
	if not tmthing.fastfalltrue and tmthing.skin == skin and (abs(thing.z - tmthing.z) <= 60*tmthing.scale)
	then
		local bottomheight = tmthing.z
		local topheight = tmthing.z + tmthing.height

		if (tmthing.eflags & MFE_VERTICALFLIP)
			local swap = bottomheight
			bottomheight = topheight
			topheight = swap
		end

		if not ((P_MobjFlip(tmthing)*(bottomheight - (thing.z + thing.height/2)) > 0)
		and (P_MobjFlip(tmthing)*(tmthing.momz - thing.momz) < 0)) then return end
			
		P_KillMobj(thing, tmthing, tmthing)
		return true
	end
end
addHook("MobjCollide", function(thing, tmthing) spikeDestroyer(thing, tmthing) end, MT_SPIKE) 		-- Destroy spikes
addHook("MobjCollide", function(thing, tmthing) spikeDestroyer(thing, tmthing) end, MT_WALLSPIKE) 	-- Destroy wallspikes

-- (Used for all animation things like setting frames, sprite2, related things)
addHook("PostThinkFrame", function(player)
	for player in players.iterate do
		-- Only continue if this is the player skin we want and the player isn't in pain
		if sanityCheck(player) and player.mo.skin ~= skin or P_PlayerInPain(player) then
			return
		end
		
		--print(player.mo.collideanim)
		--print(player.mo.collideanimtime)
		
		-- Idle animation
		if player.mo.state == S_PLAY_STND and P_IsObjectOnGround(player.mo) then
			player.mo.frame = player.mo.idletimer / 6;
			player.mo.idletimer = $ + 1
			player.mo.state = S_PLAY_STND
		end
		if player.mo.idletimer >= 29 then
			player.mo.idletimer = 0
		end
		
		-- Go to crouching state
		if not player.mo.crouchtrue then
			if not player.mo.uncrouch then
				player.mo.state = S_PLAY_HCROUCH
				player.mo.uncrouch = true
			end
		elseif player.mo.crouchtrue then
			if player.mo.uncrouch then
				player.mo.state = S_PLAY_STND
				player.mo.uncrouch = false
			end
		end
		
		-- rolling animation/etc
		--if not player.mo.divetrue then
		--	print("ROLLING")
		--end
		
		-- airgrab animation/etc
		--if not player.mo.airgrabtrue then
		--	print("AIRGRAB")
		--end

		-- if currently doing the quake attack
		if not player.mo.fastfalltrue then
			-- TODO: Stop attack momentum when you hit something? So 'bounce' off?
			if player.mo.fastfalltime <= 1 then
				P_SetObjectMomZ(player.mo,-11*FRACUNIT,false)
				P_InstaThrust(player.mo, player.drawangle, 22*FRACUNIT) --TODO: should I use drawangle here?
				-- TODO: P_Earthquake(mobj_t inflictor, mobj_t source, fixed_t radius)
			end
			player.mo.fastfalltime = $ + 1
			
			if player.mo.fastfallsfx then
				S_StartSound(player.mo, sfx_s3k3c)
				player.mo.fastfallsfx = false
			end
			
			if not P_IsObjectOnGround(player.mo) then
				player.mo.state = S_PLAY_HFASTFALL
			end
		end -- end of quake attack effects
		
		-- if the collision animation variable is true then play a short 'attack' animation
		if (player.mo.aekickstate == 1 and not P_IsObjectOnGround(player.mo)) then
			-- TOOD: maybe use a 'frame' timer to perform an 'animation'
			player.mo.state = S_PLAY_AESTH_KICK1A
			player.mo.aekickstate = 2
		-- If in the air still after the animation counter is done then reset animation to a roll state
		--elseif (player.mo.collideanim and not P_IsObjectOnGround(player.mo)) and player.mo.collideanimtime > 25 then
		--	player.mo.state = S_PLAY_ROLL
		--	player.mo.collideanimtime = 0
		--	player.mo.collideanim = false
		-- otherwise if you're grounded reset everything
		--elseif player.mo.collideanim and (P_IsObjectOnGround(player.mo) or player.mo.fastfalltrue) then
		--	player.mo.collideanimtime = 0
		--	player.mo.collideanim = false
		elseif P_IsObjectOnGround(player.mo) then
			player.mo.aekickstate = 0
		end
		
		if (player.mo.aekickstate == 2) and not P_IsObjectOnGround(player.mo) then
			P_SpawnGhostMobj(player.mo)
			player.mo.aeghost = $ + 1
			if player.mo.aeghost >= 10 then
				player.mo.aeghost = 0
				player.mo.aekickstate = 0
			end
		end
			
		
		
	end
end)

--=========Credit: Bloopsdoob=========
-- Occurs when player dies
addHook("MobjDeath", function(h, i, s, d) --these arguments can be renamed as needed, though i and s are unused.
    if h and h.valid
    and h.player and h.player.valid
    and h.skin == "honey"
        if d == DMG_FIRE
            h.deathtype = 1
        elseif d == DMG_ELECTRIC
            h.deathtype = 2
        elseif d == DMG_NUKE
            h.deathtype = 3
        end
    end
end, MT_PLAYER)

addHook("MobjThinker", function(cloth)
    if cloth and cloth.valid
        if not P_IsObjectOnGround(cloth)
            cloth.rollangle = $ + ANGLE_22h
        end
    end
end, MT_HONEY_NUKE_GIB)
--====================================

-- Check if battlemod was loaded or not
local BMLoaded = false
addHook("ThinkFrame", function()
	-- Check each frame if script was loaded or not
	if not BMLoaded and CBW_Battle
	then
		-- Mark that BattleMod was loaded
		BMLoaded = true
		local B = CBW_Battle
		
		local ActionState1 = function(player)
			local mo = player.mo
			-- Halt momentum
			P_InstaThrust(mo, mo.angle, 0)
			P_SetObjectMomZ(mo, P_MobjFlip(mo)*mo.scale, 0)
			-- Spin
			player.pflags = ($|PF_SPINNING)&~PF_JUMPED
			mo.state = S_PLAY_ROLL
			P_SpawnThokMobj(player) --aesthetic
							
			if player.actiontime == 10 -- Set time until next state
				player.actionstate = 2
				player.actiontime = 0
			end
		end
		
		local ActionState2 = function(player)
			local mo = player.mo
			-- Create projectile 
			-- Returns true once every 4 tics
			if player.actiontime%3 == 0 then
				P_SPMAngle(mo,MT_REDRING,mo.angle,0)
			end
				
			-- Recoil
			P_SetObjectMomZ(mo, P_MobjFlip(mo)*mo.scale*2,0)
			P_InstaThrust(mo, mo.angle+ANGLE_180,mo.scale*6)
			B.ApplyCooldown(player,TICRATE*2)
			mo.state = S_PLAY_FALL
			player.pflags = $ & ~PF_SPINNING
			player.drawangle = mo.angle
				
			-- Reset player state
			if player.actiontime == 20 then
				player.actionstate = 0
			end
		end
		
		-- The action function
		local RingBarrage = function(mo, doaction)
			local player = mo.player
			
			-- Red ring action text
			player.actiontext = "Unnamed Attack"
			-- Amount of rings required to do said attack
			player.actionrings = 5
			--If a string is defined, this will display below the first string, next to the flag icon.
			-- (so therefore this is defined as the dropflag atack name)
			player.action2text = nil
			--this is also the amount of rings required for the 2nd action
			-- 0 means unused.
			player.action2rings = 0
			
			if not B.CanDoAction(player) then
				-- default actionstate
				player.actionstate = 0
				return
			end
			
			-- Behaviour of action's neutral state
			if player.actionstate == 0 then
				-- 1: Action key was pressed and action can be performed
				if doaction == 1 then
					player.actionstate = 1
					player.actiontime = 0
					B.PayRings(player)
					S_StartSound(mo, sfx_s3k3c) -- spend rings sound
				end
			end
			
			-- The startup lag state
			if player.actionstate == 1 then
				-- increment actiontime to keep track of how long we've been in this state
				player.actiontime = $ + 1
				
				-- Perform state 1 action
				ActionState1(player)
			end
			
			-- In this state the character is currently performing their action
			if player.actionstate == 2 then
				-- increment actiontime to keep track of how long we've been in this state
				player.actiontime = $ + 1
				
				ActionState2(player)
			end	
		end -- End of action function
		
		-- Making character resistant to attacks while his move is being used.
		local PriorityFunc = function(player)
			if player.actionstate == 1 then
				B.SetPriority(player, 1, 3, nil, 1, 3, "midair charge spin")
				/* Arguments
				1: Player properties to modify
				2: Attack priority. Higher priorities will pierce through stronger enemy defense.
				3: Defense priority. Higher priorities will resist enemy attacks.
				4: "Special" priority condition. If returns true, uses arguments 5 & 6 to determine attack/defense priority for this frame.
					Functions used must first be stored via CBW_Battle.AddPriorityFunction(stringname,function)
					Argument #4 must then refer to the given "stringname" from AddPriorityFunction(). Do not use the priority function itself as argument #4!
				5: Special attack, if argument 4's function returns true.
				6: Special defense, if argument 4's function returns true.
				7: Attack text to display in the console when an opponent takes damage from a player-to-player collision
					(i.e. "Sonic's midair charge spin hit Knuckles"
				*/
			end
		end -- End of resistance function
		
		-- Assign the above functions to the character
		CBW_Battle.SkinVars[skin] = {
			special = RingBarrage,
			func_priority_ext = PriorityFunc
		}
				/* All of the vanilla six character's specials can also be assigned to custom characters too. The functions for them are as follows:
			Sonic: CBW_Battle.Action.SuperSpinJump
			Tails: CBW_Battle.Action.RoboMissile
			Knuckles: CBW_Battle.Action.Dig
			Amy: CBW_Battle.Action.PikoSpin
			Fang: CBW_Battle.Action.BombThrow
			Metal Sonic: CBW_Battle.Action.EnergyAttack
		*/
	end
end)